#pragma once

// C++ system level
#include <map>
#include <ctime>
#include <string>
#include <functional>

// ESP32 specific
#include "nvs.h"

// project specific
#include "SpiderLib/NumberTypes.hpp"



namespace SiliconTorch {

  namespace Metrics {

    typedef std::function<void()> Incrementer;

    u32 uptime_ms();  // rolls over after about 136 years

    class CounterMetric {
      public:
        CounterMetric(const std::string& shortName, const std::string& fullName);

        const std::string fullName;
        const std::string shortName;

        bool dirty = false;

        u64 getValue();

        void inc();
        void dec();

        void load(nvs_handle_t* nvs);       // load value from NVS
        void save(nvs_handle_t* nvs);       // saves only if enoughg time elapsed
        void forceSave(nvs_handle_t* nvs);  // saves value immediately

        u64& operator++();    // pre-increment
        u64& operator--();    // pre-decrement

        u64 operator++(int);  // post-increment
        u64 operator--(int);  // post-decrement

      private:

        u64 value = 0;

    };

    class Metrics {
      public:
        Metrics(const std::string& nameSpace);

        void registerMetric(const std::string& shortName, const std::string& fullName);

        /*
         * returns the actual value or 0 for unknown counters
         * also supports getting the special bootCnt, uptime and lifetime metrics
        */
        u64 getCounterValue(const std::string& shortName);

        Incrementer generateMetricIncrementer(const std::string& shortName);

        void saveData();
        void publishData();

        const std::string nameSpace;

        std::string genDeviceTopic(const char *suffix);

      private:

        u32 saveInterval      = 60*60;        // seconds
        u32 publishInterval   = 60;           // seconds

        u32 lastSaved         = uptime_ms();  // unix timestamp
        u32 lastPublished     = uptime_ms();  // unix timestamp

        u32 tickPeriod        = 3000;         // milli seconds

        u64 lifeTime          = 0;            // milli seconds
        u64 lifeTimeSaved     = 0;            // milli seconds

        u32 bootCnt           = 0;
        u32 bootCntSaveUptime = 60;           // seconds

        u64 birth             = 0;            // unix timestamp of first device activation

        void tickTask();
        void saveBootCntTask();

        std::map<std::string, CounterMetric*> counters;

        nvs_handle_t nvs;

    };

  }
}
